# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import codecs
from lxml import etree
import simplejson
import yaml

from .category import Category
from .etree_utils import get_node_at_path
from .logic_parser import lexpr
from .normalization import normalize_token
from .semantic_rule import SemanticRule

class SemanticIndex(object):
    def __init__(self, contents):
        # Input might be a string containing a filename, or a list of rules.
        if isinstance(contents, str) and contents != '':
            self.rules = load_semantic_rules(contents)
        elif isinstance(contents, list):
            self.rules = contents
        else:
            self.rules = []

    def get_relevant_rules(self, rule_pattern):
        """
        Given a rule pattern (that is, a SemanticRule with several features
        specified, but no semantics associated), it searches for relevant
        rules with the same features but with associated semantics.
        """
        relevant_rules = []
        for rule in self.rules:
            if rule.match(rule_pattern):
                relevant_rules.append(rule)
        return relevant_rules

    def get_semantic_representation(self, ccg_tree, tokens):
        rule_pattern = make_rule_pattern_from_ccg_node(ccg_tree, tokens)
        # Obtain the semantic template.
        relevant_rules = self.get_relevant_rules(rule_pattern)
        if not relevant_rules and len(ccg_tree) == 2:
            return None
        elif not relevant_rules:
            semantic_template = build_default_template(rule_pattern, ccg_tree)
            semantic_rule = None
        else:
            semantic_rule = relevant_rules.pop()
            semantic_template = semantic_rule.semantics
        # Apply template to relevant (current, child or children) CCG node(s).
        if len(ccg_tree) == 0:
            base = rule_pattern.attributes.get('base')
            surf = rule_pattern.attributes.get('surf')
            assert base and surf, 'The current CCG node should contain attributes ' \
              + '"base" and "surf". CCG node: {0}\nrule_pattern attributes: {1}'\
              .format(etree.tostring(ccg_tree, pretty_print=True),
                      rule_pattern.attributes)
            predicate_string = base if base != '*' else surf
            predicate = lexpr(predicate_string)
            semantics = semantic_template(predicate).simplify()
            # Assign coq types.
            if semantic_rule != None and 'coq_type' in semantic_rule.attributes:
                coq_types = semantic_rule.attributes['coq_type']
                ccg_tree.set('coq_type',
                  'Parameter {0} : {1}.'.format(predicate_string, coq_types))
            else:
                ccg_tree.set('coq_type', "")
        elif len(ccg_tree) == 1:
            predicate = lexpr(ccg_tree[0].get('sem'))
            semantics = semantic_template(predicate).simplify()
            # Assign coq types.
            ccg_tree.set('coq_type', ccg_tree[0].attrib.get('coq_type', ""))
        else:
            var_paths = semantic_rule.attributes.get('var_paths', [[0], [1]])
            semantics = semantic_template
            coq_types_list = []
            for path in var_paths:
                child_node = get_node_at_path(ccg_tree, path)
                child_semantics = lexpr(child_node.get('sem'))
                semantics = semantics(child_semantics).simplify()
                child_coq_types = child_node.get('coq_type', None)
                if child_coq_types is not None and child_coq_types != "":
                    coq_types_list.append(child_coq_types)
            if coq_types_list:
                ccg_tree.set('coq_type', ' ||| '.join(coq_types_list))
        return semantics

def get_attributes_from_ccg_node_recursively(ccg_tree, tokens):
    """
    Copies attributes from children node into the current node,
    to make them accessible in constant time.
    """
    if 'child' in ccg_tree.attrib:
        attributes = ccg_tree.attrib
        for i, child in enumerate(ccg_tree):
            child_attributes = get_attributes_from_ccg_node_recursively(child, tokens)
            for name, value in child_attributes.items():
                attributes['child' + str(i) + '_' + name] = value
    else:
        attributes = ccg_tree.attrib
        token_id = ccg_tree.get('terminal')
        token_node = find_node_by_id(token_id, tokens)
        node_id = attributes.get('id')
        attributes.update(token_node.attrib)
        # Restore the node ID, that has been overwritten by the previous update.
        attributes['id'] = node_id
    return attributes

def make_rule_pattern_from_ccg_node(ccg_tree, tokens):
    attributes = get_attributes_from_ccg_node_recursively(ccg_tree, tokens)
    category = ccg_tree.get('category')
    assert category, 'There should be a non-empty category attribute in {0}'\
      .format(etree.tostring(ccg_tree, pretty_print=True))
    semantics = None
    rule_pattern = SemanticRule(category, semantics, attributes)
    return rule_pattern

def find_node_by_id(node_id, xml_tree):
    nodes = xml_tree.xpath('.//descendant-or-self::*[@id="%s"]' % node_id)
    if not nodes:
        raise(ValueError('It should have found a span for id {0}'.format(node_id)))
    return nodes[0]

def load_semantic_rules(fn):
    semantic_rules = []
    loaded = None
    with codecs.open(fn, 'r', 'utf-8') as infile:
        loaded = yaml.load(infile)
    if not loaded: raise ValueError("couldn't load file: " + fn)

    for attributes in loaded:
        # Compulsory fields.
        category = attributes['category']
        semantics = lexpr(attributes['semantics'])
        del attributes['category'], attributes['semantics']
        for attr_name, attr_val in attributes.items():
          if attr_name.endswith('base') or attr_name.endswith('surf'):
            attributes[attr_name] = normalize_token(attr_val)
        new_semantic_rule = \
          SemanticRule(category, semantics, attributes)
        semantic_rules.append(new_semantic_rule)
    return semantic_rules

def build_default_template(rule_pattern, ccg_tree):
    category = rule_pattern.category
    if len(ccg_tree) == 0:
        num_arguments = category.get_num_args()
    elif len(ccg_tree) == 1:
        category2 = Category(ccg_tree.get('category'))
        num_arguments = category.get_num_args() - category2.get_num_args()
    variable_names = ['x' + str(i) for i in range(num_arguments)]
    if not variable_names:
        template_string = r'\P.P'
    else:
        template_string = r'\E O.O'
    template = lexpr(template_string)
    return template



from typing import Callable, List, Dict, Union, Tuple
from multiprocessing import Pool
import numpy
import depccg._parsing
import time
import math
from depccg.types import Token, CombinatorResult, ScoringResult
from depccg.tree import ScoredTree
from depccg.cat import Category


def _chunks(list_, num_chunks):
    splits = math.ceil(len(list_) / max(num_chunks, 1))
    for i in range(0, len(list_), splits):
        yield list_[i:i + splits]


def _binarize(indices, length):
    result = numpy.ones(length, dtype=numpy.bool)
    result[indices] = 0
    return result


def _type_check(doc, score_results, categories):
    many_sentences = (
        isinstance(doc, list)
        and isinstance(doc[0], list)
        and isinstance(doc[0][0], Token)
    )
    many_scores = (
        isinstance(score_results, list)
        and isinstance(score_results[0], ScoringResult)
    )

    if (
        many_sentences != many_scores
        or many_sentences and len(doc) != len(score_results)
    ):
        raise RuntimeError(
            'doc and score_results are not compatible types.')

    if not many_sentences:
        score_results = [score_results]
        doc = [doc]

    num_tags = len(categories)
    for tokens, (tag_scores, dep_scores) in zip(doc, score_results):
        if num_tags != tag_scores.shape[1]:
            raise RuntimeError(
                ("all inputs to depccg.parsing.run must contain scores for"
                 " the equal number of categories as the `categories` list.")
            )
        num_tokens = len(tokens)
        expected_tag_score = (num_tokens, num_tags)
        expected_dep_score = (num_tokens, num_tokens + 1)
        if (
            expected_tag_score != tag_scores.shape
            or expected_dep_score != dep_scores.shape
        ):
            raise RuntimeError(
                ('invalid shape of input matrices:\n'
                 f'Expected P_tag: {expected_tag_score}, P_dep: {expected_dep_score}\n'
                 f'Actual P_tag: {tag_scores.shape}, P_dep: {dep_scores.shape}')
            )

    return doc, score_results


def apply_category_filters(
    doc: Union[Token, List[List[Token]]],
    score_results: Union[ScoringResult, List[ScoringResult]],
    categories: List[Category],
    category_dict: Dict[str, List[Category]],
    large_negative_value: float = -10e+32,
) -> Union[Tuple[List[Token], ScoringResult], Tuple[List[List[Token]], List[ScoringResult]]]:

    doc, score_results = _type_check(doc, score_results, categories)

    category_ids = {
        cat: index for index, cat in enumerate(categories)
    }

    category_dict = {
        word: _binarize(
            [category_ids[cat] for cat in cats],
            score_results[0].tag_scores.shape[1]
        )
        for word, cats in category_dict.items()
    }

    for tokens, (tag_scores, _) in zip(doc, score_results):
        for index, token in enumerate(tokens):
            if token.word in category_dict:
                tag_scores[index, category_dict[token.word]
                           ] = large_negative_value

    return doc, score_results


def run(
    doc: Union[Token, List[List[Token]]],
    score_results: Union[ScoringResult, List[ScoringResult]],
    categories: List[Category],
    root_categories: List[Category],
    binary_fun: Callable[[Category, Category], List[CombinatorResult]],
    unary_fun: Callable[[Category], List[CombinatorResult]],
    unary_penalty: float = 0.1,
    beta: float = 0.00001,
    use_beta: bool = True,
    pruning_size: int = 50,
    nbest: int = 1,
    max_step: int = 10000000,
    max_length: int = 250,
    processes: int = 2,
    max_chunk_size: int = 20,
) -> List[List[ScoredTree]]:

    doc, score_results = _type_check(doc, score_results, categories)

    args = (
        categories,
        binary_fun,
        unary_fun,
        root_categories
    )

    kwargs = {
        'num_tags': score_results[0][0].shape[1],
        'unary_penalty': unary_penalty,
        'beta': beta,
        'use_beta': use_beta,
        'pruning_size': pruning_size,
        'nbest': nbest,
        'max_step': max_step,
        'max_length': max_length
    }

    if len(doc) <= max_chunk_size:

        results = depccg._parsing.run(
            doc,
            score_results,
            *args,
            **kwargs,
        )

    else:
        chunks = _chunks(list(zip(doc, score_results)), processes)
        with Pool(processes) as pool:
            tasks = []
            for chunk_index, chunk in enumerate(chunks):
                doc_, score_results_ = zip(*chunk)
                task = pool.apply_async(
                    depccg._parsing.run,
                    args=(list(doc_), list(score_results_)) + args,
                    kwds={**kwargs, 'process_id': chunk_index},
                )
                tasks.append(task)

            while not all(task.ready() for task in tasks):
                time.sleep(1)

            results = [
                result
                for task in tasks
                for result in task.get()
            ]

    return results




from overrides import overrides
from allennlp_models.structured_prediction.dataset_readers.penn_tree_bank import \
    PennTreeBankConstituencySpanDatasetReader
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from depccg.tools.reader import read_auto
from nltk.tree import Tree


def ccg_to_nltk_tree(tree):
    def rec(node):
        if node.is_leaf:
            cat = node.cat
            children = [Tree('XX', [node.word])]  # add dummy pos non-terminal
        else:
            cat = node.cat
            children = [rec(child) for child in node.children]
        return Tree(str(cat), children)
    return rec(tree)


@DatasetReader.register("ccgbank_as_ptb")
class CCGBankAsConstituencyReader(PennTreeBankConstituencySpanDatasetReader):
    @overrides
    def _read(self, file_path):
        for _, _, ccg_tree in read_auto(file_path):
            tree = ccg_to_nltk_tree(ccg_tree)
            pos_tags = [x[1]
                        for x in tree.pos()] if self._use_pos_tags else None
            yield self.text_to_instance(
                tree.leaves(), pos_tags, tree)




import logging

logger = logging.getLogger(__name__)


GLOBAL_LANG_NAME = 'en'


def set_global_language_to(lang: str) -> None:
    global GLOBAL_LANG_NAME
    logger.info('Setting the global language config to: %s', lang)
    GLOBAL_LANG_NAME = lang


def get_global_language() -> str:
    return GLOBAL_LANG_NAME



# C&C NLP tools
# Copyright (c) Universities of Edinburgh, Oxford and Sydney
# Copyright (c) James R. Curran
#
# This software is covered by a non-commercial use licence.
# See LICENCE.txt for the full text of the licence.
#
# If LICENCE.txt is not included in this distribution
# please email candc@it.usyd.edu.au to obtain a copy.

import sys
import re
import logging
import argparse
import tempfile
import os
import subprocess
from pathlib import Path

from depccg.tools.reader import read_auto
from depccg import utils
from depccg.printer.auto import auto_flattened_of

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                    level=logging.INFO)
logger = logging.getLogger(__name__)


IGNORE_RULES = r"""
rule_id 7
rule_id 11
rule_id 12
rule_id 14
rule_id 15
rule_id 16
rule_id 17
rule_id 51
rule_id 52
rule_id 56
rule_id 91
rule_id 92
rule_id 95
rule_id 96
rule_id 98
conj 1 0
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 0
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 2
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 3
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 6
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 9
((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 6
((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_} 1 6
(((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_}/NP{W}<3>){_} 1 6
(S[X]{Y}/S[X]{Y}<1>){_} 1 13
(S[X]{Y}/S[X]{Y}<1>){_} 1 5
(S[X]{Y}/S[X]{Y}<1>){_} 1 55
((S[X]{Y}/S[X]{Y}){Z}\(S[X]{Y}/S[X]{Y}){Z}<1>){_} 2 97
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 4
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 93
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 8
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 94
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 18
been ((S[pt]{_}\NP{Y}<1>){_}/(S[ng]{Z}<2>\NP{Y*}){Z}){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[adj]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[pt]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[adj]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[ng]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
going ((S[ng]{_}\NP{Y}<1>){_}/(S[to]{Z}<2>\NP{Y*}){Z}){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[to]{Z}<2>\NP{Y*}){Z}){_} 1 0
Here (S[adj]{_}\NP{Y}<1>){_} 1 0
# this is a dependency Julia doesn't have but looks okay
from (((NP{Y}\NP{Y}<1>){_}/(NP{Z}\NP{Z}){W}<3>){_}/NP{V}<2>){_} 1 0
"""

IGNORE = {tuple(rule.split())
          for rule in IGNORE_RULES.split('\n') if not rule.startswith('#')}


def die(msg):
    logger.error(msg)
    sys.exit(1)


def get_deps_from_auto(auto_file):
    candc_dir = os.environ.get('CANDC', None)
    if not candc_dir:
        die('did not find C&C parser at CANDC environmental variable.')
    CANDC_DIR = Path(candc_dir).resolve()
    GENERATE = CANDC_DIR / 'bin' / 'generate'
    MARKEDUP = CANDC_DIR / 'src' / 'data' / 'ccg' / 'cats' / 'markedup'
    CATS = CANDC_DIR / 'src' / 'data' / 'ccg' / 'cats'
    if not GENERATE.exists():
        logger.error(
            'Currently the evalution script requires C&C parser compiled from its source.')
        die('expected: $CANDC/bin/generate')
    elif not MARKEDUP.exists() or not CATS.exists:
        logger.error('The C&C directory is not configured expectedly.')
        die('expected: $CANDC/src/data/ccg/cats/markedup')

    tmp = tempfile.mktemp()
    print(tmp)
    with open(tmp, 'w') as f:
        for _, _, tree in read_auto(auto_file):
            print(auto_flattened_of(tree), file=f)

    command = f'{GENERATE} -j {CATS} {MARKEDUP} {tmp}'
    proc = subprocess.Popen(command,
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    results, error = proc.communicate()
    if len(error.decode('utf-8')) > 0:
        die(
            f'caught error in running $CANDC/bin/generate: {error.decode("utf-8")}')

    lines = iter(results.decode('utf-8').split('\n'))
    deps, udeps = set(), set()
    rule_ids = {}
    line = next(lines)
    while line != '':
        line = next(lines)

    for line in lines:
        line = line.strip()
        if len(line) == 0:
            # If 0, no dependencies for this sentence - probably a conversion script error.
            parsed = len(rule_ids) > 0
            yield parsed, deps, udeps, rule_ids
            deps, udeps = set(), set()
            rule_ids = {}
            continue
        fields = line.split()
        pred, cat, slot, arg, rule_id = fields[:5]
        pred_word = pred.rsplit('_')[0]
        arg_word = arg.rsplit('_')[0]
        if not ignore(pred_word, cat, slot, arg_word, rule_id):
            cat = strip_markup(cat)
            deps.add((pred, cat, slot, arg))
            rule_ids[(pred, cat, slot, arg)] = rule_id
            udeps.add((pred, arg))


def get_pargs(file):
    try:
        lines = open(file)
    except IOError as e:
        die(f'could not open gold_deps file ({e.strerror})')

    deps, udeps = set(), set()
    for line in lines:
        line = line.strip()
        if line.startswith('<s '):
            continue
        if line.startswith('<\\s>'):
            yield deps, udeps
            deps, udeps = set(), set()
            continue
        arg_index, pred_index, cat, slot, arg, pred = line.split()[:6]
        pred = f'{utils.normalize(pred)}_{int(pred_index) + 1}'
        arg = f'{utils.normalize(arg)}_{int(arg_index) + 1}'
        deps.add((pred, cat, slot, arg))
        udeps.add((pred, arg))
    assert len(deps) == 0 and len(udeps) == 0


DEPS_IGNORED = 0


def ignore(pred, cat, slot, arg, rule_id):
    global DEPS_IGNORED
    res = ('rule_id', rule_id) in IGNORE or \
          (cat, slot, rule_id) in IGNORE or \
          (pred, cat, slot, rule_id) in IGNORE or \
          (pred, cat, slot, arg, rule_id) in IGNORE
    DEPS_IGNORED += res
    return res


MARKUP = re.compile(r'<[0-9]>|\{[A-Z_]\*?\}|\[X\]')


def strip_markup(cat):
    cat = MARKUP.sub('', cat)
    return cat[1:-1] if cat[0] == '(' else cat


def score_deps(gold_deps, test_deps, rule_ids, verbose, relations,
               correct_relations, incorrect_relations, missing_relations):
    correct = gold_deps.intersection(test_deps)
    if verbose:
        for dep in correct:
            print('correct: %s %s %s %s %s' % (dep + (rule_ids[dep],)))
    if relations:
        for dep in correct:
            correct_relations[dep[1:3]] = correct_relations.setdefault(
                dep[1:3], 0) + 1

    incorrect = test_deps.difference(gold_deps)
    if verbose:
        for dep in incorrect:
            print('incorrect: %s %s %s %s %s' % (dep + (rule_ids[dep],)))
    if relations:
        for dep in incorrect:
            incorrect_relations[dep[1:3]] = incorrect_relations.setdefault(
                dep[1:3], 0) + 1

    missing = gold_deps.difference(test_deps)
    if verbose:
        for dep in missing:
            print('missing:   %s %s %s %s ?' % dep)
    if relations:
        for dep in missing:
            missing_relations[dep[1:3]] = missing_relations.setdefault(
                dep[1:3], 0) + 1

    if verbose:
        print()
    return len(correct), len(incorrect), len(missing)


def score_udeps(gold_deps, test_deps):
    correct = gold_deps.intersection(test_deps)
    incorrect = test_deps.difference(gold_deps)
    missing = gold_deps.difference(test_deps)
    return len(correct), len(incorrect), len(missing)


def percentage(val, total):
    return 100.0 * val / total if val else 0.0


def print_acc(name, desc, correct, total):
    acc = percentage(correct, total)
    print(f'{name:6s}: {acc:5.2f}% ({correct} of {total} {desc})')


def print_stats(name, correct, incorrect, missing):
    test = correct + incorrect
    gold = correct + missing
    prec = percentage(correct, test)
    recall = percentage(correct, gold)
    fscore = (2 * prec * recall) / (prec + recall) if prec and recall else 0.0
    print(
        f'{name[0]}p:    {prec:5.2f}% ({correct} of {test} {name} deps precision)')
    print(
        f'{name[0]}r:    {recall:5.2f}% ({correct} of {gold} {name} deps recall)')
    print(f'{name[0]}f:    {fscore:5.2f}% ({name} deps f-score)')


def print_rel_stats(relation, correct, incorrect, missing):
    relation = "%s %s" % relation
    test = correct + incorrect
    prec = percentage(correct, test)
    gold = correct + missing
    recall = percentage(correct, gold)
    fscore = (2 * prec * recall) / (prec + recall) if prec and recall else 0.0
    print(f'{relation:50s}: {prec:6.2f}% {recall:6.2f}% {fscore:6.2f}% {test:6d} {gold:6d} {correct:6d}')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('PARG_FILENAME', type=Path,
                        help='gold parg file in ccgbank')
    parser.add_argument('TEST', type=Path,
                        help='parse results in auto file format')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='produces verbose output')
    parser.add_argument('-r', '--relations', action='store_true',
                        help='produces per relation output')
    args = parser.parse_args()

    preface = f'# this file was generated by the following command(s):\n# {" ".join(sys.argv)}\n'

    nsentences, parse_failures = 0, 0
    deps_sent_correct, deps_correct, deps_incorrect, deps_missing = 0, 0, 0, 0
    udeps_sent_correct, udeps_correct, udeps_incorrect, udeps_missing = 0, 0, 0, 0
    relations_correct, relations_incorrect, relations_missing = {}, {}, {}

    TEST = get_deps_from_auto(args.TEST)
    for gold_deps, gold_udeps in get_pargs(args.PARG_FILENAME):
        parsed, test_deps, test_udeps, test_rule_ids = next(TEST)
        nsentences += 1
        if not parsed:
            parse_failures += 1
            continue

        correct, incorrect, missing = score_deps(gold_deps,
                                                 test_deps,
                                                 test_rule_ids,
                                                 args.verbose,
                                                 args.relations,
                                                 relations_correct,
                                                 relations_incorrect,
                                                 relations_missing)
        deps_correct += correct
        deps_incorrect += incorrect
        deps_missing += missing
        deps_sent_correct += incorrect == 0 and missing == 0

        correct, incorrect, missing = score_udeps(gold_udeps, test_udeps)
        udeps_correct += correct
        udeps_incorrect += incorrect
        udeps_missing += missing
        udeps_sent_correct += incorrect == 0 and missing == 0

    print(f'''{preface}
note: all these statistics are over just those sentences
      for which the parser returned an analysis, and
      dependency extraction script is successful
''')

    if args.relations:
        relations = relations_correct.copy()
        for relation, freq in relations_missing.items():
            relations[relation] = relations.get(relation, 0) + freq

        for relation, freq in sorted(relations.items()):
            print_rel_stats(relation,
                            relations_correct.get(relation, 0),
                            relations_incorrect.get(relation, 0),
                            relations_missing.get(relation, 0))
        print()

    nparsed = nsentences - parse_failures
    print_acc('cover', 'sentences evaluated  - this includes dependency extraction script errors and parse failures', nparsed, nsentences)
    print()
    print_stats('labelled', deps_correct, deps_incorrect, deps_missing)
    print_acc('lsent', 'labelled deps sentences correct',
              deps_sent_correct, nparsed)
    print()
    print_stats('unlabelled', udeps_correct, udeps_incorrect, udeps_missing)
    print_acc('usent', 'unlabelled deps sentences correct',
              udeps_sent_correct, nparsed)
    print()
    print_acc('skip', 'ignored deps (to ensure compatibility with CCGbank)',
              DEPS_IGNORED, deps_correct + deps_incorrect + DEPS_IGNORED)


if __name__ == '__main__':
    main()



from typing import Callable, List, TypeVar, Tuple, Set, Dict
from depccg.types import Combinator, CombinatorResult
from depccg.cat import Category

X = TypeVar('X')
Pair = Tuple[X, X]


def apply_rules(
    left: Category,
    right: Category,
    seen_rules: Set[Pair[Category]],
    combinators: List[Combinator],
    cache: Dict[Pair[Category], List[CombinatorResult]],
) -> List[CombinatorResult]:
    cats = (left, right)
    if cats in cache:
        return cache[cats]

    results = []
    if cats in seen_rules:
        for combinator in combinators:
            result = combinator(*cats)
            if result is not None:
                results.append(result)

    cache[cats] = results
    return results


def guess_combinator_by_triplet(
    binary_rules: Callable[[Category, Category], List[Category]],
    target: Category,
    x: Category,
    y: Category,
) -> CombinatorResult:
    for rule in binary_rules(x, y):
        if rule.cat == target:
            return rule
    return CombinatorResult(
        cat=rule.cat,
        op_string="unk",
        op_symbol="<unk>",
        head_is_left=True
    )



from typing import Union
import os
import json
import logging
import chainer

from depccg.chainer.lstm_parser_bi_fast import FastBiaffineLSTMParser
from depccg.chainer.ja_lstm_parser_bi import BiaffineJaLSTMParser


logger = logging.getLogger(__name__)


def load_chainer_tagger(
    model_path: str,
    device: int = -1
) -> Union[FastBiaffineLSTMParser, BiaffineJaLSTMParser]:

    model_file = os.path.join(model_path, 'tagger_model')
    def_file = os.path.join(model_path, 'tagger_defs.txt')

    assert os.path.exists(model_file) and os.path.exists(def_file), \
        (f'Failed in initialization. Directory "{model_path}" must contain both'
            '"tagger_model" and "tagger_defs.txt" files')

    with open(def_file) as f:
        tagger = eval(json.load(f)['model'])(model_path)

    logger.info(f'initializing supertagger with parameters at {model_file}')
    chainer.serializers.load_npz(model_file, tagger)

    if device >= 0:
        logger.info(f'sending the supertagger to gpu: {device}')
        tagger.to_gpu(device)

    return tagger



import sys
import logging

import depccg.parsing
from depccg.types import Token
from depccg.cat import Category
from depccg.printer import print_
from depccg.instance_models import load_model
from depccg.argparse import parse_args
from depccg.lang import set_global_language_to
from depccg.annotator import (
    english_annotator, japanese_annotator, annotate_XX
)
from depccg.allennlp.utils import read_params

# disable lengthy allennlp logs
logging.getLogger('filelock').setLevel(logging.ERROR)
logging.getLogger('allennlp').setLevel(logging.ERROR)
logger = logging.getLogger(__name__)


def get_annotator(args):
    if args.lang == 'en':
        if (
            args.format in ['ccg2lambda', 'jigg_xml_ccg2lambda']
            and args.annotator is None
        ):
            raise RuntimeError(
                ('Specify --annotator argument in '
                 f'using "{args.format}" output format')
            )
        return english_annotator.get(args.annotator, annotate_XX)

    elif args.lang == 'ja':
        if (
            args.format in ['ccg2lambda', 'jigg_xml_ccg2lambda']
            and not args.tokenize
        ):
            raise RuntimeError(
                ('Cannot specify --pre-tokenized '
                 f'argument using "{args.format}" output format')
            )
        if args.tokenize:
            return japanese_annotator[args.annotator]
        return annotate_XX

    raise RuntimeError(f'unsupported language: {args.lang}')


def main(args):
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
        level=logging.CRITICAL if args.silent else logging.INFO
    )

    set_global_language_to(args.lang)
    annotator_fun = get_annotator(args)
    supertagger, config = load_model(args.model, args.gpu)

    (
        apply_binary_rules,
        apply_unary_rules,
        category_dict,
        _
    ) = read_params(config.config, args)

    root_categories = [
        Category.parse(category)
        for category in args.root_cats.split('|')
    ]

    semantic_templates = (
        args.semantic_templates or config.semantic_templates
    )

    kwargs = dict(
        unary_penalty=args.unary_penalty,
        nbest=args.nbest,
        pruning_size=args.pruning_size,
        beta=args.beta,
        use_beta=not args.disable_beta,
        max_length=args.max_length,
        max_step=args.max_step,
        processes=args.num_processes,
    )

    if args.input is not None:
        input_type = open(args.input)
    elif not sys.stdin.isatty():
        input_type = sys.stdin
    else:
        # reading from keyboard
        input_type = None
        sys.stdout.flush()
        sys.stderr.flush()
        logging.getLogger().setLevel(logging.CRITICAL)

    categories = None
    while True:
        fin = [
            line for line in map(str.strip, input_type or [input()])
            if len(line) > 0
        ]
        if len(fin) == 0:
            break

        if args.input_format == 'POSandNERtagged':
            doc = [
                [
                    Token.of_piped(token)
                    for token in sent.split(' ')
                ] for sent in fin
            ]

        else:
            doc = annotator_fun(
                [
                    [word for word in sentence.split(' ')]
                    for sentence in fin
                    if len(sentence) > 0
                ],
                tokenize=args.tokenize,
            )

        logger.info("supertagging")
        score_result, categories_ = supertagger.predict_doc(
            [[token.word for token in sentence] for sentence in doc]
        )
        if categories is None:
            categories = [
                Category.parse(category) for category in categories_
            ]

        if category_dict is not None:
            doc, score_result = depccg.parsing.apply_category_filters(
                doc,
                score_result,
                categories,
                category_dict,
            )

        logger.info("parsing")
        results = depccg.parsing.run(
            doc,
            score_result,
            categories,
            root_categories,
            apply_binary_rules,
            apply_unary_rules,
            **kwargs,
        )

        print_(
            results,
            format=args.format,
            semantic_templates=semantic_templates
        )

        if input_type is None:
            sys.stdout.flush()
        else:
            break


if __name__ == '__main__':
    parse_args(main)



from typing import Dict, Optional, Tuple, Any, List
import logging
import copy

from overrides import overrides
import torch
from torch.nn.modules import Dropout
import numpy

from allennlp.common.checks import check_dimensions_match
from allennlp.data import Vocabulary
from allennlp.data.vocabulary import DEFAULT_OOV_TOKEN
from allennlp.modules import Seq2SeqEncoder, TextFieldEmbedder
from allennlp.modules.matrix_attention.bilinear_matrix_attention import BilinearMatrixAttention
from allennlp.modules import FeedForward
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator, RegularizerApplicator, Activation
from allennlp.nn.util import get_text_field_mask, get_range_vector
from allennlp.nn.util import get_device_of, masked_log_softmax
from allennlp.training.metrics import CategoricalAccuracy
from depccg.allennlp.nn.bilinear import BilinearWithBias

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _apply_head_mask(attended_arcs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    # Mask the diagonal, because the head of a word can't be itself.
    attended_arcs = attended_arcs + \
        torch.diag(attended_arcs.new(mask.size(1)).fill_(-numpy.inf))
    # Mask padded tokens, because we only want to consider actual words as heads.
    attended_arcs.masked_fill_((~mask).unsqueeze(1), -numpy.inf)
    attended_arcs.masked_fill_((~mask).unsqueeze(2), -numpy.inf)
    return attended_arcs


@Model.register("supertagger")
class Supertagger(Model):
    def __init__(
        self,
        vocab: Vocabulary,
        text_field_embedder: TextFieldEmbedder,
        encoder: Seq2SeqEncoder,
        tag_representation_dim: int,
        arc_representation_dim: int,
        tag_feedforward: FeedForward = None,
        arc_feedforward: FeedForward = None,
        dropout: float = 0.5,
        input_dropout: float = 0.5,
        head_tag_temperature: Optional[float] = None,
        head_temperature: Optional[float] = None,
        initializer: InitializerApplicator = InitializerApplicator(),
        regularizer: Optional[RegularizerApplicator] = None
    ) -> None:
        super(Supertagger, self).__init__(vocab, regularizer)

        self.text_field_embedder = text_field_embedder
        self.encoder = encoder

        encoder_dim = encoder.get_output_dim()

        self.head_arc_feedforward = arc_feedforward or FeedForward(
            encoder_dim, 1,
            arc_representation_dim,
            Activation.by_name("elu")(),
            dropout=dropout
        )
        self.child_arc_feedforward = copy.deepcopy(self.head_arc_feedforward)

        self.arc_attention = BilinearMatrixAttention(
            arc_representation_dim,
            arc_representation_dim,
            use_input_biases=True
        )

        num_labels = self.vocab.get_vocab_size("head_tags")

        self.head_tag_feedforward = tag_feedforward or FeedForward(
            encoder_dim, 1,
            tag_representation_dim,
            Activation.by_name("elu")(),
            dropout=dropout
        )
        self.child_tag_feedforward = copy.deepcopy(self.head_tag_feedforward)

        self.tag_bilinear = BilinearWithBias(
            tag_representation_dim,
            tag_representation_dim,
            num_labels
        )
        self._head_sentinel = torch.nn.Parameter(
            torch.randn([1, 1, encoder.get_output_dim()]))

        representation_dim = text_field_embedder.get_output_dim()

        check_dimensions_match(representation_dim, encoder.get_input_dim(),
                               "text field embedding dim", "encoder input dim")
        check_dimensions_match(tag_representation_dim, self.head_tag_feedforward.get_output_dim(),
                               "tag representation dim", "tag feedforward output dim")
        check_dimensions_match(arc_representation_dim, self.head_arc_feedforward.get_output_dim(),
                               "arc representation dim", "arc feedforward output dim")
        self._input_dropout = Dropout(input_dropout)
        self._attachment_scores = CategoricalAccuracy()
        self._tagging_accuracy = CategoricalAccuracy()
        self.head_tag_temperature = head_tag_temperature
        self.head_temperature = head_temperature
        initializer(self)

    @overrides
    def forward(
        self,  # type: ignore
        words: Dict[str, torch.LongTensor],
        weight: torch.Tensor,
        metadata: List[Dict[str, Any]],
        head_tags: torch.LongTensor = None,
        head_indices: torch.LongTensor = None
    ) -> Dict[str, torch.Tensor]:

        # pylint: disable=arguments-differ
        embedded_text_input = self.text_field_embedder(words)
        embedded_text_input = self._input_dropout(embedded_text_input)

        mask = get_text_field_mask(words)
        encoded_text = self.encoder(embedded_text_input, mask)

        batch_size, _, encoding_dim = encoded_text.size()

        head_sentinel = self._head_sentinel.expand(batch_size, 1, encoding_dim)
        # Concatenate the head sentinel onto the sentence representation.
        encoded_text = torch.cat([head_sentinel, encoded_text], 1)
        mask = torch.cat([mask.new_ones(batch_size, 1), mask], 1)
        if head_indices is not None:
            head_indices = torch.cat(
                [head_indices.new_zeros(batch_size, 1), head_indices], 1)
        if head_tags is not None:
            head_tags = torch.cat(
                [head_tags.new_zeros(batch_size, 1), head_tags], 1)

        # shape (batch_size, sequence_length, arc_representation_dim)
        head_arc_representation = self.head_arc_feedforward(encoded_text)
        child_arc_representation = self.child_arc_feedforward(encoded_text)

        # shape (batch_size, sequence_length, tag_representation_dim)
        head_tag_representation = self.head_tag_feedforward(encoded_text)
        child_tag_representation = self.child_tag_feedforward(encoded_text)
        # shape (batch_size, sequence_length, sequence_length)
        attended_arcs = self.arc_attention(
            head_arc_representation,
            child_arc_representation
        )

        if head_indices is not None and head_tags is not None:
            loss, normalised_arc_logits, normalised_head_tag_logits = self._construct_loss(
                head_tag_representation=head_tag_representation,
                child_tag_representation=child_tag_representation,
                attended_arcs=attended_arcs,
                head_indices=head_indices,
                head_tags=head_tags,
                mask=mask,
                weight=weight
            )

            normalised_arc_logits = _apply_head_mask(
                normalised_arc_logits, mask)
            tag_mask = self._get_unknown_tag_mask(mask, head_tags)
            self._attachment_scores(
                normalised_arc_logits[:, 1:], head_indices[:, 1:], mask[:, 1:])
            self._tagging_accuracy(
                normalised_head_tag_logits[:, 1:], head_tags[:, 1:], tag_mask[:, 1:])
            predicted_heads, predicted_head_tags = None, None
        else:
            attended_arcs = _apply_head_mask(attended_arcs, mask)
            # Compute the heads greedily.
            # shape (batch_size, sequence_length)
            _, predicted_heads = attended_arcs.max(dim=2)

            # Given the greedily predicted heads, decode their dependency tags.
            # shape (batch_size, sequence_length, num_head_tags)
            head_tag_logits = self._get_head_tags(
                head_tag_representation,
                child_tag_representation,
                predicted_heads
            )
            _, predicted_head_tags = head_tag_logits.max(dim=2)

            loss, normalised_arc_logits, normalised_head_tag_logits = self._construct_loss(
                head_tag_representation=head_tag_representation,
                child_tag_representation=child_tag_representation,
                attended_arcs=attended_arcs,
                head_indices=predicted_heads.long(),
                head_tags=predicted_head_tags.long(),
                mask=mask,
                weight=weight
            )
            normalised_arc_logits = _apply_head_mask(
                normalised_arc_logits, mask
            )

        output_dict = {
            "heads": normalised_arc_logits,
            "head_tags": normalised_head_tag_logits,
            "loss": loss,
            "mask": mask,
            "words": [meta["words"] for meta in metadata],
        }

        if predicted_heads is not None and predicted_head_tags is not None:
            output_dict['predicted_heads'] = predicted_heads[:, 1:]
            output_dict['predicted_head_tags'] = predicted_head_tags[:, 1:]
            output_dict = self.decode(output_dict)
        return output_dict

    def decode(self, output_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        head_tags = output_dict.pop("head_tags")
        # discard a sentinel token and padding and unknown tags
        head_tags = head_tags[:, 1:, 2:]
        heads = output_dict.pop("heads")
        heads = heads[:, 1:]
        output_dict["head_tags"] = head_tags.cpu().detach().numpy()
        output_dict["heads"] = heads.cpu().detach().numpy()

        if 'predicted_heads' in output_dict:
            output_dict['predicted_heads'] = (
                output_dict['predicted_heads'].cpu().detach().numpy()
            )

        if 'predicted_head_tags' in output_dict:
            output_dict['predicted_head_tags'] = (
                output_dict['predicted_head_tags'].cpu().detach().numpy()
            )

        output_dict.pop('loss')
        return output_dict

    def _construct_loss(
        self,
        head_tag_representation: torch.Tensor,
        child_tag_representation: torch.Tensor,
        attended_arcs: torch.Tensor,
        head_indices: torch.Tensor,
        head_tags: torch.Tensor,
        mask: torch.Tensor,
        weight: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        float_mask = mask.float()
        tag_mask = self._get_unknown_tag_mask(mask, head_tags)

        batch_size, sequence_length, _ = attended_arcs.size()
        # shape (batch_size, 1)
        range_vector = get_range_vector(
            batch_size, get_device_of(attended_arcs)).unsqueeze(1)
        # shape (batch_size, sequence_length, sequence_length)
        if self.head_temperature:
            attended_arcs /= self.head_temperature
        normalised_arc_logits = (
            masked_log_softmax(attended_arcs, mask) *
            float_mask.unsqueeze(2) * float_mask.unsqueeze(1)
        )

        # shape (batch_size, sequence_length, num_head_tags)
        if self.head_tag_temperature:
            attended_arcs /= self.head_tag_temperature
        head_tag_logits = self._get_head_tags(
            head_tag_representation, child_tag_representation, head_indices
        )
        normalised_head_tag_logits = (
            masked_log_softmax(
                head_tag_logits, tag_mask.unsqueeze(-1)) * tag_mask.float().unsqueeze(-1)
        )
        # index matrix with shape (batch, sequence_length)
        timestep_index = get_range_vector(
            sequence_length, get_device_of(attended_arcs))
        child_index = timestep_index.view(1, sequence_length).expand(
            batch_size, sequence_length).long()
        # shape (batch_size, sequence_length)
        arc_loss = normalised_arc_logits[
            range_vector, child_index, head_indices
        ]
        tag_loss = normalised_head_tag_logits[
            range_vector, child_index, head_tags
        ]
        tag_loss *= (head_tags > 1).float()
        # We don't care about predictions for the symbolic ROOT token's head,
        # so we remove it from the loss.
        arc_loss = arc_loss[:, 1:] * weight
        tag_loss = tag_loss[:, 1:] * weight

        # The number of valid positions is equal to the number of unmasked elements minus
        # 1 per sequence in the batch, to account for the symbolic HEAD token.
        valid_positions = mask.sum() - batch_size

        arc_nll = -arc_loss.sum() / valid_positions.float()
        tag_nll = -tag_loss.sum() / valid_positions.float()
        loss = arc_nll + tag_nll
        return loss, normalised_arc_logits, normalised_head_tag_logits

    def _get_head_tags(
        self,
        head_tag_representation: torch.Tensor,
        child_tag_representation: torch.Tensor,
        head_indices: torch.Tensor
    ) -> torch.Tensor:

        batch_size = head_tag_representation.size(0)
        # shape (batch_size,)
        range_vector = get_range_vector(
            batch_size, get_device_of(head_tag_representation)).unsqueeze(1)

        # This next statement is quite a complex piece of indexing, which you really
        # need to read the docs to understand. See here:
        # https://docs.scipy.org/doc/numpy-1.13.0/reference/arrays.indexing.html#advanced-indexing
        # In effect, we are selecting the indices corresponding to the heads of each word from the
        # sequence length dimension for each element in the batch.

        # shape (batch_size, sequence_length, tag_representation_dim)
        selected_head_tag_representations = head_tag_representation[range_vector, head_indices]
        selected_head_tag_representations = selected_head_tag_representations.contiguous()
        # shape (batch_size, sequence_length, num_head_tags)
        head_tag_logits = self.tag_bilinear(
            selected_head_tag_representations, child_tag_representation
        )

        return head_tag_logits

    def _get_unknown_tag_mask(
        self,
        mask: torch.LongTensor,
        head_tags: torch.LongTensor
    ) -> torch.LongTensor:
        oov = self.vocab.get_token_index(DEFAULT_OOV_TOKEN, 'head_tags')
        new_mask = mask.detach()
        oov_mask = head_tags.eq(oov).long()
        new_mask = new_mask * (1 - oov_mask)
        return new_mask

    @overrides
    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        dependency = self._attachment_scores.get_metric(reset)
        tagging = self._tagging_accuracy.get_metric(reset)
        harmonic_mean = (2 * dependency * tagging) / (dependency + tagging)
        scores = {
            'dependency': dependency,
            'tagging': tagging,
            'harmonic_mean': harmonic_mean,
        }
        return scores



from depccg.tree import Tree
from depccg.utils import normalize, denormalize


def auto_of(tree: Tree) -> str:
    """tree string in auto format commonly used in English CCGBank.

    Args:
        tree (Tree): tree object

    Returns:
        str: tree string in the auto format
    """

    def rec(node):
        if node.is_leaf:
            cat = node.cat
            word = denormalize(node.word)
            pos = node.token.get('pos', 'POS')
            return f'(<L {cat} {pos} {pos} {word} {cat}>)'
        else:
            cat = node.cat
            children = ' '.join(rec(child) for child in node.children)
            num_children = len(node.children)
            head_is_left = 0 if node.head_is_left else 1
            return f'(<T {cat} {head_is_left} {num_children}> {children} )'

    return rec(tree)


def auto_flattened_of(tree: Tree) -> str:
    """tree string in flattened version of auto format.
    This is mainly used for evaluation only.

    Args:
        tree (Tree): tree object

    Returns:
        str: tree string in the flattened auto format
    """

    def rec(node):
        if node.is_leaf:
            cat = node.cat
            word = normalize(node.word).replace('/', '\\/')
            pos = node.token.get('pos', 'POS')
            return f'(<L *** {cat} {pos} {word}>\n)'
        else:
            cat = node.cat
            children = '\n'.join(rec(child) for child in node.children)
            num_children = len(node.children)
            head_is_left = 0 if node.head_is_left else 1
            return f'(<T *** {cat} * {head_is_left} {num_children}>\n{children}\n)'

    return f'###\n{rec(tree)}\n'


def auto_extended_of(tree: Tree) -> str:
    """tree string in extended version of auto format, used by C&C.

    Args:
        tree (Tree): tree object

    Returns:
        str: tree string in the extended auto format
    """

    def rec(node):
        if node.is_leaf:
            cat = node.cat
            word = denormalize(node.word)
            token = node.token
            lemma = token.get('lemma', 'XX')
            pos = token.get('pos', 'XX')
            entity = token.get('entity', 'XX')
            chunk = token.get('chunk', 'XX')
            return f'(<L {cat} {word} {lemma} {pos} {entity} {chunk} {cat}>)'
        else:
            cat = node.cat
            children = ' '.join(rec(child) for child in node.children)
            num_children = len(node.children)
            head_is_left = 0 if node.head_is_left else 1
            rule = node.op_string
            return f'(<T {cat} {rule} {head_is_left} {num_children}> {children} )'

    return rec(tree)



import pytest
from depccg.cat import Category, Atom, Functor, TernaryFeature, UnaryFeature


en_categories = [
    (Category.parse(text.strip()), text.strip())
    for text in open('tests/cats.txt')
]

ja_categories = [
    (Category.parse(text.strip()), text.strip())
    for text in open('tests/cats.ja.txt')
]


@pytest.mark.parametrize("result, expect", en_categories)
def test_parse_many_en(result, expect):
    assert str(result) == expect


@pytest.mark.parametrize("result, expect", en_categories)
def test_parse_many_ja(result, expect):
    assert str(result) == expect


def test_parse():
    assert Category.parse("NP") == Atom("NP")
    assert Category.parse("(NP)") == Atom("NP")
    assert Category.parse("S/NP") == Functor(Atom("S"), '/', Atom("NP"))
    assert Category.parse("(S/NP)") == Functor(Atom("S"), '/', Atom("NP"))
    assert Category.parse(
        "S[dcl]/NP") == Functor(Atom("S", UnaryFeature("dcl")), '/', Atom("NP"))

    # japanese categories
    assert Category.parse("NP[case=nc,mod=nm,fin=t]") == Atom(
        "NP", TernaryFeature(("case", "nc"), ("mod", "nm"), ("fin", "t")))
    assert Category.parse("S[mod=nm,form=attr,fin=t]") == Atom(
        "S", TernaryFeature(("mod", "nm"), ("form", "attr"), ("fin", "t")))

    # punctuations
    assert Category.parse(',') == Atom(',')
    assert Category.parse('.') == Atom('.')
    assert Category.parse(';') == Atom(';')
    assert Category.parse(':') == Atom(':')
    assert Category.parse('LRB') == Atom('LRB')
    assert Category.parse('RRB') == Atom('RRB')
    assert Category.parse('conj') == Atom('conj')
    assert Category.parse('*START*') == Atom('*START*')
    assert Category.parse('*END*') == Atom('*END*')


def test_binop():
    S = Category.parse("S")
    NP = Category.parse("NP")

    assert S / NP == Category.parse("S/NP")
    assert S | NP == Category.parse("S\\NP")
    assert S | (S / NP) == Category.parse("S\\(S/NP)")
    assert (S | NP) / NP == Category.parse("(S\\NP)/NP")

    assert S / NP == "S/NP"
    assert S | NP == "S\\NP"
    assert S | (S / NP) == "S\\(S/NP)"
    assert (S | NP) / NP == "(S\\NP)/NP"

    assert "S/NP" == S / NP
    assert "S\\NP" == S | NP
    assert "S\\(S/NP)" == S | (S / NP)
    assert "(S\\NP)/NP" == (S | NP) / NP



from depccg.tree import Tree


def ptb_of(tree: Tree) -> Tree:
    """PTB-style string of a CCG tree

    Args:
        tree (Tree): tree object

    Returns:
        str: tree string in the PTB style
    """

    def rec(node):
        if node.is_leaf:
            cat = node.cat
            word = node.word
            return f'({cat} {word})'
        else:
            cat = node.cat
            children = ' '.join(rec(child) for child in node.children)
            return f'({cat} {children})'

    return f'(ROOT {rec(tree)})'



from setuptools import Extension, setup, find_packages
import sys
import os
import contextlib
import subprocess


try:
    from Cython.Build import build_ext
except ImportError:
    sys.exit("""\
Could not import Cython, which is required to build depccg extension modules.
Please install cython and numpy prior to installing depccg.\
""")

try:
    import numpy
except ImportError:
    sys.exit("""\
Could not import numpy, which is required to build the extension modules.
Please install cython and numpy prior to installing depccg.\
""")

here = os.path.abspath(os.path.dirname(__file__))


install_requires = [
    line.strip() for line in open(
        os.path.join(here, 'requirements.txt'))
]

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        sys.path.insert(0, new_dir)
        yield
    finally:
        del sys.path[0]
        os.chdir(old_dir)


def clean():
    depccg_dir = os.path.join(here, 'depccg')
    for name in os.listdir(depccg_dir):
        file_path = os.path.join(depccg_dir, name)
        if any(file_path.endswith(ext) for ext in ['.so', '.cpp', '.c']):
            os.unlink(file_path)
    with chdir('c'):
        subprocess.call(["make", "clean"], env=os.environ)


def generate_cpp(options):
    options = ' '.join(options)
    options = f'OPTIONS={options}'
    with chdir('c'):
        p = subprocess.call(["make", options], env=os.environ)
        if p != 0:
            raise RuntimeError('Running cythonize failed')


COMPILE_OPTIONS = [
    '-O3',
    '-Wall',
    '-std=c++14'
]

LINK_OPTIONS = []

if sys.platform == 'darwin':
    COMPILE_OPTIONS.append('-stdlib=libc++')
    LINK_OPTIONS.append('-lc++')
    # g++ (used by unix compiler on mac) links to libstdc++ as a default lib.
    # See: https://stackoverflow.com/questions/1653047/avoid-linking-to-libstdc
    LINK_OPTIONS.append('-nodefaultlibs')

ext_modules = [
    Extension(
        'depccg.morpha',
        ['depccg/morpha.pyx'],
        language='c++',
        extra_compile_args=COMPILE_OPTIONS,
        extra_link_args=LINK_OPTIONS + ['c/morpha.o'],
        include_dirs=['.', 'c']
    ),
    Extension(
        'depccg._parsing',
        ['depccg/parsing.pyx'],
        language='c++',
        extra_compile_args=COMPILE_OPTIONS,
        extra_link_args=LINK_OPTIONS,
        include_dirs=[numpy.get_include(), '.', 'depccg'],
    )
]


if len(sys.argv) > 1 and sys.argv[1] == 'clean':
    clean()
else:
    generate_cpp([])

    setup(
        name="depccg",
        version="2.0.0",  # NOQA
        description='A parser for natural language based on combinatory categorial grammar',
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='Masashi Yoshikawa',
        author_email='yoshikawa@tohoku.ac.jp',
        url='https://github.com/masashi-y/depccg',
        license='MIT License',
        packages=find_packages(),
        package_data={'depccg': ['models/*']},
        scripts=['bin/depccg_en', 'bin/depccg_ja'],
        install_requires=install_requires,
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
        zip_safe=False,
        cmdclass={'build_ext': build_ext},
        ext_modules=ext_modules
    )



from typing import Optional, NamedTuple, Callable, List
from pathlib import Path
import re
import numpy

from depccg.cat import Category

dunder_pattern = re.compile("__.*__")
protected_pattern = re.compile("_.*")


class Token(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, item):
        if dunder_pattern.match(item) or protected_pattern.match(item):
            return super().__getattr__(item)
        return self[item]

    def __repr__(self):
        res = super().__repr__()
        return f'Token({res})'

    @classmethod
    def of_piped(cls, string: str) -> 'Token':
        # WORD|POS|NER or WORD|LEMMA|POS|NER
        # or WORD|LEMMA|POS|NER|CHUCK
        items = string.split('|')
        if len(items) == 5:
            word, lemma, pos, entity, chunk = items
        elif len(items) == 4:
            word, lemma, pos, entity = items
            chunk = 'XX'
        else:
            assert len(items) == 3
            word, pos, entity = items
            lemma = 'XX'
            chunk = 'XX'

        return Token(
            word=word,
            lemma=lemma,
            pos=pos,
            entity=entity,
            chunk=chunk
        )

    @classmethod
    def of_word(cls, word: str) -> 'Token':
        return Token(
            word=word,
            lemma='XX',
            pos='XX',
            entity='XX',
            chunk='XX'
        )


class CombinatorResult(NamedTuple):
    cat: Category
    op_string: str
    op_symbol: str
    head_is_left: bool


class ScoringResult(NamedTuple):
    tag_scores: numpy.ndarray
    dep_scores: numpy.ndarray


Combinator = Callable[[Category, Category], Optional[CombinatorResult]]

ApplyBinaryRules = Callable[..., List[CombinatorResult]]
ApplyUnaryRules = Callable[..., List[CombinatorResult]]


class GrammarConfig(NamedTuple):
    apply_binary_rules: ApplyBinaryRules
    apply_unary_rules: ApplyUnaryRules


class ModelConfig(NamedTuple):
    framework: str
    name: str
    url: str
    config: Path
    semantic_templates: Path



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from lxml import etree
from multiprocessing import Pool
from multiprocessing import Lock
import os
import sys

from nltk.sem.logic import LogicalExpressionException
from .ccg2lambda_tools import assign_semantics_to_ccg
from .semantic_index import SemanticIndex
from .logic_parser import lexpr
from .nltk2normal import remove_true

logger = logging.getLogger(__name__)

SEMANTIC_INDEX = None
GOLD_TREES = True
NBEST = 0
SENTENCES = None
kMaxTasksPerChild = None
lock = Lock()


def parse(ccg, templates, nbest=0, ncores=3):
    global SEMANTIC_INDEX
    global SENTENCES
    global NBEST
    NBEST = nbest

    if not os.path.exists(templates):
        print('File does not exist: {0}'.format(templates))
        sys.exit(1)

    logger.info(templates)
    SEMANTIC_INDEX = SemanticIndex(templates)
    SENTENCES = ccg.findall('.//sentence')
    print(SENTENCES)
    sentence_inds = range(len(SENTENCES))
    sem_nodes_lists = semantic_parse_sentences(sentence_inds, ncores)
    assert len(sem_nodes_lists) == len(SENTENCES), \
        'Element mismatch: {0} vs {1}'.format(
            len(sem_nodes_lists), len(SENTENCES))
    logging.info('Adding XML semantic nodes to sentences...')
    formulas_list = []
    for sentence, (sem_nodes, orig_formulas) in zip(SENTENCES, sem_nodes_lists):
        formulas = []
        for formula in orig_formulas:
            try:
                formulas.append(str(remove_true(lexpr(formula))))
            except LogicalExpressionException:
                formulas.append(formula)
        formulas_list.append(formulas)
        sentence.extend(sem_nodes)
    logging.info('Finished adding XML semantic nodes to sentences.')

    root_xml_str = serialize_tree(ccg)
    return root_xml_str, formulas_list


def semantic_parse_sentences(sentence_inds, ncores=1):
    if ncores <= 1:
        sem_nodes_lists = semantic_parse_sentences_seq(sentence_inds)
    else:
        sem_nodes_lists = semantic_parse_sentences_par(sentence_inds, ncores)
    results = [([etree.fromstring(s) for s in sem_nodes], formulas)
               for sem_nodes, formulas in sem_nodes_lists]
    return results


def semantic_parse_sentences_par(sentence_inds, ncores=3):
    pool = Pool(processes=ncores, maxtasksperchild=kMaxTasksPerChild)
    results = pool.map(semantic_parse_sentence, sentence_inds)
    pool.close()
    pool.join()
    return results


def semantic_parse_sentences_seq(sentence_inds):
    results = []
    for sentence_ind in sentence_inds:
        result = semantic_parse_sentence(sentence_ind)
        results.append(result)
    return results


def semantic_parse_sentence(sentence_ind):
    """
    `sentence` is an lxml tree with tokens and ccg nodes.
    It returns an lxml semantics node.
    """
    global lock
    sentence = SENTENCES[sentence_ind]
    sem_nodes = []
    formulas = []

    tree_indices = [int(sentence.get('gold_tree', '0')) + 1]
    if NBEST != 1:
        tree_indices = get_tree_indices(sentence, NBEST)
    for tree_index in tree_indices:
        sem_node = etree.Element('semantics')
        try:
            sem_tree = assign_semantics_to_ccg(
                sentence, SEMANTIC_INDEX, tree_index)
            filter_attributes(sem_tree)
            sem_node.extend(sem_tree.xpath('.//descendant-or-self::span'))
            sem_node.set('status', 'success')
            sem_node.set('ccg_id',
                         sentence.xpath('./ccg[{0}]/@id'.format(tree_index))[0])
            sem_node.set('root',
                         sentence.xpath('./ccg[{0}]/@root'.format(tree_index))[0])
            formulas.append(sem_tree.attrib['sem'])
        except Exception as e:
            sem_node.set('status', 'failed')
            # from pudb import set_trace; set_trace()
            sentence_surf = ' '.join(sentence.xpath('tokens/token/@surf'))
            lock.acquire()
            logging.error('An error occurred: {0}\nSentence: {1}\nTree XML:\n{2}'.format(
                e, sentence_surf,
                etree.tostring(sentence, encoding='utf-8', pretty_print=True).decode('utf-8')))
            lock.release()
            # print('x', end='', file=sys.stdout)
            formulas.append('FAILED!')
        sem_nodes.append(sem_node)
    sem_nodes = [etree.tostring(sem_node) for sem_node in sem_nodes]
    return sem_nodes, formulas


def get_tree_indices(sentence, nbest):
    num_ccg_trees = int(sentence.xpath('count(./ccg)'))
    if nbest < 1:
        nbest = num_ccg_trees
    return list(range(1, min(nbest, num_ccg_trees) + 1))


keep_attributes = set(['id', 'child', 'sem', 'type'])


def filter_attributes(tree):
    if 'coq_type' in tree.attrib and 'child' not in tree.attrib:
        sem_type = \
            tree.attrib['coq_type'].lstrip('["Parameter ').rstrip('."]')
        if sem_type:
            tree.attrib['type'] = sem_type
    attrib_to_delete = [
        a for a in tree.attrib.keys() if a not in keep_attributes]
    for a in attrib_to_delete:
        del tree.attrib[a]
    for child in tree:
        filter_attributes(child)
    return


def serialize_tree(tree):
    tree_str = etree.tostring(
        tree, xml_declaration=True, encoding='utf-8', pretty_print=True)
    return tree_str



from typing import Dict, Any, List

from overrides import overrides
from allennlp.common.util import JsonDict
from allennlp.data import DatasetReader, Instance
from allennlp.models import Model
from allennlp.predictors.predictor import Predictor
from allennlp.data.vocabulary import DEFAULT_OOV_TOKEN, DEFAULT_PADDING_TOKEN


@Predictor.register('supertagger-predictor')
class SupertaggerPredictor(Predictor):
    def __init__(self, model: Model, dataset_reader: DatasetReader) -> None:
        super().__init__(model, dataset_reader)

    def predict(self, sentence: str) -> JsonDict:
        return self.predict_json({"sentence": sentence})

    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        return self._dataset_reader.text_to_instance(json_dict["sentence"])

    @overrides
    def predict_instance(self, instance: Instance) -> JsonDict:
        outputs = self._model.forward_on_instance(instance)
        [result] = self._make_json([outputs])
        return result

    @overrides
    def predict_batch_instance(self, instances: List[Instance]) -> List[JsonDict]:
        outputs = self._model.forward_on_instances(instances)
        return self._make_json(outputs)

    def _make_json(self, output_dicts: List[Dict[str, Any]]) -> List[JsonDict]:
        all_categories = self._model.vocab.get_index_to_token_vocabulary(
            'head_tags')
        all_categories = [token for _, token in sorted(all_categories.items())]
        categories, paddings = all_categories[2:], all_categories[:2]
        assert all(padding in [DEFAULT_PADDING_TOKEN,
                               DEFAULT_OOV_TOKEN] for padding in paddings)
        for output_dict in output_dicts:
            length = len(output_dict['words'].split(' '))
            output_dict['categories'] = categories
            head_tags = output_dict["head_tags"][:length, :]
            assert head_tags.shape[-1] == len(
                categories), f"{head_tags.shape[-1]}, {len(all_categories)}"
            heads = output_dict["heads"][:length, :length + 1]
            output_dict["predicted_head_tags"] = [
                all_categories[i] for i in output_dict["predicted_head_tags"]
            ]
            output_dict["predicted_heads"] = output_dict["predicted_heads"].tolist()
            output_dict["head_tags"] = head_tags.flatten().astype(
                float).tolist()
            output_dict["heads"] = heads.flatten().astype(float).tolist()
            output_dict["head_tags_shape"] = list(head_tags.shape)
            output_dict["heads_shape"] = list(heads.shape)
            output_dict.pop("mask")
        return output_dicts



from typing import Dict, Tuple, Optional
import tarfile
import logging
from pathlib import Path
from collections import defaultdict

from depccg.types import GrammarConfig, ModelConfig
from depccg.chainer.supertagger import load_chainer_tagger
from depccg.allennlp.supertagger import load_allennlp_tagger
from depccg.lang import get_global_language
from depccg.grammar import en, ja

logger = logging.getLogger(__name__)

MODEL_DIRECTORY = Path(__file__).parent / 'models'


SEMANTIC_TEMPLATES: Dict[str, Path] = {
    'en': MODEL_DIRECTORY / 'semantic_templates_en_event.yaml',
    'ja': MODEL_DIRECTORY / 'semantic_templates_ja_event.yaml'
}

GRAMMARS: Dict[str, GrammarConfig] = {
    'en': GrammarConfig(
        en.apply_binary_rules,
        en.apply_unary_rules,
    ),
    'ja': GrammarConfig(
        ja.apply_binary_rules,
        ja.apply_unary_rules,
    )
}

MODELS: Dict[str, ModelConfig] = {
    'en': ModelConfig(
        'chainer',
        'tri_headfirst',
        '1mxl1HU99iEQcUYhWhvkowbE4WOH0UKxv',
        MODEL_DIRECTORY / 'config_en.jsonnet',
        SEMANTIC_TEMPLATES['en'],
    ),
    'en[elmo]': ModelConfig(
        'allennlp',
        'lstm_parser_elmo',
        '1UldQDigVq4VG2pJx9yf3krFjV0IYOwLr',
        MODEL_DIRECTORY / 'config_en.jsonnet',
        SEMANTIC_TEMPLATES['en'],
    ),
    'en[rebank]': ModelConfig(
        'allennlp',
        'lstm_parser_char_rebanking',
        '1Az840uCW8QuAkNCZq_Y8VOkW5j0Vtcj9',
        MODEL_DIRECTORY / 'config_rebank.jsonnet',
        SEMANTIC_TEMPLATES['en'],
    ),
    'en[elmo_rebank]': ModelConfig(
        'allennlp',
        'lstm_parser_elmo_rebanking',
        '1deyCjSgCuD16WkEhOL3IXEfQBfARh_ll',
        MODEL_DIRECTORY / 'config_rebank.jsonnet',
        SEMANTIC_TEMPLATES['en'],
    ),
    'ja': ModelConfig(
        'chainer',
        'ja_headfinal',
        '1bblQ6FYugXtgNNKnbCYgNfnQRkBATSY3',
        MODEL_DIRECTORY / 'config_ja.jsonnet',
        SEMANTIC_TEMPLATES['ja'],
    )
}


def _lang_and_variant(model: str):
    if '[' in model and ']' in model:
        assert model[-1] == ']'
        return model[:-1].split('[')
    return model, None


def _get_model_name(variant: Optional[str]) -> str:
    lang = get_global_language()
    if variant is None:
        return lang
    return f'{lang}[{variant}]'


AVAILABLE_MODEL_VARIANTS = defaultdict(list)
for model in MODELS:
    lang, variant = _lang_and_variant(model)
    AVAILABLE_MODEL_VARIANTS[lang].append(variant)


def download(lang: str, variant: Optional[str]) -> None:
    config = MODELS[f'{lang}[{variant}]' if variant else lang]

    from google_drive_downloader import GoogleDriveDownloader as gdd
    logging.info(f'start downloading from {config.url}')
    filename = (MODEL_DIRECTORY / config.name).with_suffix('.tar.gz')
    gdd.download_file_from_google_drive(
        file_id=config.url,
        dest_path=filename,
        unzip=False,
        overwrite=True
    )

    if config.framework == 'chainer':
        logging.info('extracting files')
        tf = tarfile.open(filename)
        tf.extractall(MODEL_DIRECTORY)
    logging.info('finished')


def load_model_directory(
    variant: Optional[str]
) -> Tuple[Path, ModelConfig]:
    config = MODELS[_get_model_name(variant)]
    model_path = MODEL_DIRECTORY / config.name
    if config.framework == 'allennlp':
        model_path = model_path.with_suffix('.tar.gz')
    if not model_path.exists():
        if variant is None:
            variant = ''
        lang = get_global_language()
        raise RuntimeError(
            ('please download the model by doing '
             f'\'depccg_{lang} download {variant}\'.')
        )
    return model_path, config


def model_is_available(model_name: str) -> bool:
    return model_name in MODELS.keys()


def load_model(variant: Optional[str], device: int = -1):
    model_path, config = load_model_directory(variant)
    if config.framework == 'allennlp':
        supertagger = load_allennlp_tagger(model_path, device)
    elif config.framework == 'chainer':
        supertagger = load_chainer_tagger(model_path, device)
    else:
        lang = get_global_language()
        raise KeyError(
            ('unsupported model for language '
             f'({lang}): {variant}')
        )
    return supertagger, config
